package Reports;

=head1 NAME

Reports - Perl extension for generating generic reports in Koha

=cut

use strict;
use CGI;
use C4::Context;

=head1 VERSION

  $Id:$

=cut

=head1 DESCRIPTION

Module for querying and stocking Virtual Bookshelves

   1. can contain a list of items, a list of biblioitems, or a list of biblios
   2. can have an arbitrary name, and will have a unique numerical identifier
   3. will have arbitrary metadata (properties) associated with it
	  * Sharing information (private, only visible by the owner of the
	    Reports; shared with a group of patrons; public, viewable by anybody)
	  * Special circulation rules - Do not return to home branch, do not
	    circulate, reduced loan time (ie 3 day loan)
	  * Search query term - if the Reports is the result of a query, the
	    query itself can be stored with the list of books that resulted
          * Creation date - useful for 'retiring' a stale cached query result
          * Access information - who has "write" or "read" access to the Reports.
	  * Searchable - If a patron can perform a search query on the contents
	    of this Reports


Patrons typically will only use "biblioitem" bookshelves, and will not need to
be presented with the differences between biblioitem and item bookshelves.


Some uses for VirtualBookshelves

   1. Cache search results for faster response on popular searches
   2. Name search results so that patrons can pull up saved searches
   3. Creation of sub-collections within a library or branch
   4. replacing "itemtypes" field... this would allow an individual item to be
	a member of more than one itemtype
   5. store a patron's reading record (if he chooses to store such data)
   6. store a patron's "To be read" list
   7. A teacher of a course could add a list of books to a Reports for his course
	and ask that those items be marked non-circulating so students always
	have access to them at the library.
	  * The teacher creates the list of materials that she wants to be
	    non-circulating (or reduced to 3-day loan) and marks them as such
	  * A librarian receives a notice that a Reports requires her attention.
	    He can pull up a list of the contents of the Reports, the owner of
	    the Reports, and the reason the owner is requesting this change in
	    circulation rules. The librarian can approve or deny the request.
	  * Optionally, create an access flag that grants teachers the right to
	    put items on modified circulation shelves without librarian
	    intervention.


=cut


my $OPERATORS= {
    'EQUALS' => { 'name' => 'is equal to', 'order' => 1 },
    'CONTAINS' => { 'name' => 'contains', 'order' => 3  },
    'GT' => { 'name' => '&gt;', 'order' => 5  },
    'LT' => { 'name' => '&lt;', 'order' => 6  },
    'GE' => { 'name' => '&ge;', 'order' => 7  },
    'LE' => { 'name' => '&le;', 'order' => 8  },
};



# This section contains hard coded report query info.  Normally this stuff
# would be stored in a database table, of course.

my $QC;

$QC->{1} = { 'name' => 'Weird and Complicated', 'query' => [ 'AND', 'AND', 'OR', 'CONTAINS', 'Title', 'Ginn', 'CONTAINS', 'Title', 'Jack', 'OR', 'CONTAINS', 'Author', 'Hall', 'OR', 'CONTAINS', 'Author', 'Berger', 'NOT', 'CONTAINS', 'Author', 'gold', 'AND', 'GT', 'Barcode', '10000', 'LT', 'Barcode', '24000' ] };
$QC->{2} = { 'name' => 'Simple Title Search', 'query' => [ 'CONTAINS', 'Title', 'nisga\'a' ] };
$QC->{3} = { 'name' => 'Limit by Publication Year', 'query' => [ 'AND', 'CONTAINS', 'Author', 'A', 'AND', 'GT', 'Publication Year', '1990', 'NOT', 'GT', 'Publication Year', '1995' ] };
$QC->{4} = { 'name' => 'Issues Query', 'query' => [ 'NOT', 'CONTAINS', 'Issued To', 'john' ] };

my $QV;
$QV->[1] = [ 'Title', 'Subtitle', 'Author', 'Barcode' ];
$QV->[2] = [ 'Title', 'Author', 'Itemtype', 'Dewey' ];
$QV->[3] = [ 'Title', 'Subtitle', 'Publication Year' ];
$QV->[4] = [ 'IssuedTo', 'CardNumber', 'Borrower Surname', 'Borrower Firstname', 'Barcode', 'Due Date', 'Title' ];


my $variabletable = {
    'Title' => 'biblio.title',
    'Author' => 'biblio.author',
    'Barcode' => 'items.barcode',
    'Subtitle' => 'bibliosubtitle.subtitle',
    'Publication Year' => 'biblioitems.publicationyear',
    'Dewey' => 'biblioitems.dewey',
    'Itemtype' => 'itemtypes.description',
    'IssuedTo' => 'issues.borrowernumber',
    'IssuedItem' => 'issues.itemnumber',
    'CardNumber' => 'borrowers.cardnumber',
    'Borrower Surname' => 'borrowers.surname',
    'Borrower Firstname' => 'borrowers.firstname',
    'Branch' => 'items.holdingbranch',
    'Barcode' => 'items.barcode',
    'Due Date' => 'issues.date_due',
    'Date Last Seen' => 'items.datelastseen',
};

my $variablestable = {
    'BIBLIOGRAPHIC' => {
	'Title' => 'biblio.title',
	'Author' => 'biblio.author',
	'Barcode' => 'items.barcode',
	'Subtitle' => 'bibliosubtitle.subtitle',
	'Publication Year' => 'biblioitems.publicationyear',
	'Dewey' => 'biblioitems.dewey',
	'Itemtype' => 'itemtypes.description',
	'Branch' => 'items.holdingbranch',
	'Barcode' => 'items.barcode',
	'Date Last Seen' => 'items.datelastseen',
    },
    'BORROWERS' => {
	'CardNumber' => 'borrowers.cardnumber',
	'Borrower Surname' => 'borrowers.surname',
	'Borrower Firstname' => 'borrowers.firstname',
    },
    'ISSUES' => {
	'Due Date' => 'issues.date_due',
	'Issuing Branch' => 'issues.issuingbranch',
	'Returned Date' => 'issues.returndate',
	'Last Renew Date' => 'issues.lastrenewdate',
	'# of Renewals' => 'issues.renewals',
    },
    'RESERVES' => {
	'Reserve Date' => 'issues.reservedate',
    }
}

=head1 METHODS

=head2 C<new()>

Base constructor for the class.

  my $Reports=Reports->new(56);
      will load bookReports 56.
  my $Reports=Reports->new(-name => 'Fiction');
  my $Reports=Reports->new('Fiction');
      will load the internal 'Fiction' Reports
  my $Reports=Reports->new('Favourite Books', 'sjohnson');
  my $Reports=Reports->new(-name => 'Favourite Books', -owner => 'sjohnson');
      will load sjohnson's "Favourite Books" bookReports
  
  Any of the last four invocations will create a new Reports with the name and
  owner given if one doesn't already exist.


=cut

sub new {
    my $self = {};
    $self->{ID}	= undef;
    $self->{NAME}=undef;
    $self->{QUERY}=undef;
    $self->{VARIABLES}=undef;
    $self->{COLUMNHEADERS}=undef;

    if (@_) {
	my $dbh=C4::Context->dbh();
	shift;
	if ($#_ == 0) {
	    my $dbh=C4::Context->dbh();
	    $self->{ID}=shift;
	    my $sth=$dbh->prepare("select name,criteria,variables from reports where id=".$self->{ID});
	    $sth->execute;
	    my ($name, $query, $variables) = $sth->fetchrow;
	    @{$self->{QUERY}} = split(/\0/, $query);
	    @{$self->{VARIABLES}} = split(/\0/, $variables);
	    foreach my $field (@{$self->{VARIABLES}}) {
		foreach my $variable (keys %$variabletable) {
		    if ($variabletable->{$variable} eq $field) {
			push @{$self->{COLUMNHEADERS}}, $variable;
		    }
		}
	    }
	    $self->{NAME}=$name;
	} elsif ($#_) {
	    my ($name,$query,$variables);
	    if ($_[0] =~/^-/) {
		my %params=@_;
		$name=$params{name};
		$query=$params{query};
		$variables=$params{variables};
	    } else {
		$name=shift;
		$query=shift;
		$variables=shift;
	    }
	}
    }
    bless($self);
    return $self;
}

sub updatequery {
    my $self=shift;
    my $query=shift;
    (return) unless ($query->param('setquery'));
    #my @array= @{$self->{QUERY}};
    my @array;
    foreach my $param ($query->param) {
	if ($param=~/element-(\d+)/) {
	    my $elementcounter=$1;
	    my $value=$query->param($param);
	    $array[$elementcounter]=$value;
	}
    }
    undef $self->{QUERY};
    foreach (@array) {
	push @{$self->{QUERY}}, $_;
    }
    my @temparray;
    my %newlocations;
    my $counter=0;
    foreach (@array) {
	if (defined($_)) {
	    push @temparray, $_;
	} else {
	    for (my $i=$counter+1; $i<$#array; $i++) {
		if (defined($newlocations{$i})) {
		    $newlocations{$i}=$newlocations{$i}-1;
		} else {
		    $newlocations{$i}=$i-1;
		}
	    }
	}
	$counter++;
    }
    undef @array;
    foreach (@temparray) {
	push @array, $_;
    }
    foreach my $param ($query->param) {
	if ($param=~/add-(\d+)/) {
	    my $origelementcounter=$1;
	    my $elementcounter=$origelementcounter;
	    if (defined($newlocations{$origelementcounter})) {
		$elementcounter=$newlocations{$origelementcounter};
	    }
	    my @temparray=@array[0...$elementcounter-1];
	    push @temparray, 'AND', 'CONTAINS', 'Title', 'EDIT NEW TERM HERE';
	    push @temparray, @array[$elementcounter...$#array];
	    @array=@temparray;
	    undef @temparray;
	    #my @temparray=@{$self->{QUERY}}[0...$origelementcounter-1];
	    #push @temparray, 'NOT';
	    #push @temparray, @{$self->{QUERY}}[$origelementcounter...$#array];
	    #$self->{QUERY}=\@temparray;
	    for (my $i=$origelementcounter; $i<$#{$self->{QUERY}}; $i++) {
		if (defined($newlocations{$i})) {
		    $newlocations{$i}=$newlocations{$i}+4;
		} else {
		    $newlocations{$i}=$i+4;
		}
	    }
	} elsif ($param=~/addnot-(\d+)/) {
	    my $origelementcounter=$1;
	    my $elementcounter=$origelementcounter;
	    if (defined($newlocations{$origelementcounter})) {
		$elementcounter=$newlocations{$origelementcounter};
	    }
	    my @temparray=@array[0...$elementcounter-1];
	    push @temparray, 'NOT';
	    push @temparray, @array[$elementcounter...$#array];
	    @array=@temparray;
	    undef @temparray;
	    #my @temparray=@{$self->{QUERY}}[0...$origelementcounter-1];
	    #push @temparray, 'NOT';
	    #push @temparray, @{$self->{QUERY}}[$origelementcounter...$#array];
	    #$self->{QUERY}=\@temparray;
	    for (my $i=$origelementcounter; $i<$#{$self->{QUERY}}; $i++) {
		if (defined($newlocations{$i})) {
		    $newlocations{$i}=$newlocations{$i}+1;
		} else {
		    $newlocations{$i}=$i+1;
		}
	    }
	} elsif ($param=~/del-(\d+)/) {
	    my $origelementcounter=$1;
	    my $parent=$self->findparent($origelementcounter);
	    while ($newlocations{$parent}==-1) {
		$parent=$self->findparent($parent);
	    }
	    my $elementcounter=$origelementcounter;
	    if (defined($newlocations{$origelementcounter})) {
		$elementcounter=$newlocations{$origelementcounter};
	    }
	    @array=@array[0...$elementcounter-1,$elementcounter+3...$#array];
	    for (my $i=$origelementcounter+1; $i<$#{$self->{QUERY}}; $i++) {
		if (defined($newlocations{$i})) {
		    $newlocations{$i}=$newlocations{$i}-3;
		} else {
		    $newlocations{$i}=$i-3;
		}
	    }
	    if ($array[$elementcounter-1] eq 'NOT') {
		if ($elementcounter-1 == $#array) {
		    @array=@array[0...$elementcounter-2];
		} elsif ($elementcounter-1 > 0) {
		    @array=@array[0...$elementcounter-2,$elementcounter...$#array];
		} else {
		    @array=@array[$elementcounter...$#array];
		}
		for (my $i=$origelementcounter; $i<$#{$self->{QUERY}}; $i++) {
		    if (defined($newlocations{$i})) {
			$newlocations{$i}=$newlocations{$i}-1;
		    } else {
			$newlocations{$i}=$i-1;
		    }
		}
	    }
	    if ($parent>-1) {
		for (my $i=$parent+1; $i<$#{$self->{QUERY}}; $i++) {
		    if (defined($newlocations{$i})) {
			$newlocations{$i}=$newlocations{$i}-1;
		    } else {
			$newlocations{$i}=$i-1;
		    }
		}
		$parent=$newlocations{$parent} if defined($newlocations{$parent});
		$newlocations{$parent}=-1;
		@array=@array[0...$parent-1,$parent+1...$#array];
	    }
	}
    }
    undef $self->{QUERY};
    $self->{QUERY}=\@array;
    $self->savequery();
}


sub savequery {
    my $self=shift;
    my $query=join("\0", @{$self->{QUERY}});
    my $variables=join("\0", @{$self->{VARIABLES}});
    my $name=$self->{NAME};
    my $id=$self->{ID};
    my $dbh=C4::Context->dbh();
    my $sth=$dbh->prepare("select * from reports where id=$id");
    $sth->execute;
    if ($sth->rows) {
	$sth=$dbh->prepare("update reports set criteria=?, variables=?, name=? where id=?");
	$sth->execute($query, $variables, $name, $id);
    } else {
	$sth=$dbh->prepare("insert into reports (id, criteria, variables, name) values (?, ?, ?, ?)");
	$sth->execute($id, $query, $variables, $name);
    }

}

sub printarray {
    my @array=@_;
    my $string;
    my $elementcounter=0;
    foreach (@array) {
	$string.=$elementcounter++.": $_ ";
    }
    return;
}

sub findparent {
    my $self=shift;
    my $parentof=shift;
    my $elementcounter=$#{@{$self->{QUERY}}};
    my @holder;
    foreach my $element (reverse @{$self->{QUERY}}) {
	if ($element eq 'AND') {
	    if (pop @holder eq "ELEMENT-$parentof" or pop @holder eq "ELEMENT-$parentof") {
		return $elementcounter;
	    }
	    push @holder, "ELEMENT-$elementcounter";
	} elsif ($element eq 'OR') {
	    if (pop @holder eq "ELEMENT-$parentof" or pop @holder eq "ELEMENT-$parentof") {
		return $elementcounter;
	    }
	    push @holder, "ELEMENT-$elementcounter";
	} elsif ($element eq 'NOT') {
	   # push @holder, "ELEMENT-$elementcounter";
	} elsif ($OPERATORS->{$element}) {
	    pop @holder;
	    pop @holder;
	    push @holder, "ELEMENT-$elementcounter";
	} else {
	    push @holder, "ELEMENT-$elementcounter";
	}
	$elementcounter--;
    }
}

sub variabledropdownbox {
    my $variable=shift;
    my $elementcounter=shift;
    my $validvariables=ValidVariables();
    my $element="<select name=element-$elementcounter>\n";
    foreach my $validvariable (sort keys %$validvariables) {
	my $field=$validvariables->{$validvariable}->{field};
	my $selected='';
	$validvariable eq $variable and $selected=' selected';
	$element.="<option value='$validvariable'$selected>$validvariable\n";
    }
    $element.="</select>\n";
    return $element;
}

sub comparisondropdownbox {
    my $comparison=shift;
    my $elementcounter=shift;
    my $element="<select name=element-$elementcounter>\n";
    foreach my $op (sort { $OPERATORS->{$a}->{order} <=> $OPERATORS->{$b}->{order} } keys %$OPERATORS) {
	my $selected='';
	$op eq $comparison and $selected=' selected';
	my $opname=$OPERATORS->{$op}->{name};
	$element.="<option value=$op$selected>$opname\n";
    }
    $element.="</select>\n";
    return $element;
}

sub editvariables {
    my $self=shift;
    my $query=shift;
    my $id=$self->{ID};
    my $dbh=C4::Context->dbh();
    if ($query->param('updatevariables')) {
	my @variables;
	foreach my $param ($query->param) {
	    if ($param=~/variable-(\d+)/) {
		my $variable=$1;
		$variables[$variable]=$query->param($param);
	    }
	}
	foreach my $param ($query->param) {
	    if ($param=~/delvariable-(\d*)/) {
		my $variable=$1;
		$variables[$variable]='';
	    }
	}
	my @temparray;
	foreach (@variables) {
	    $_ and push @temparray, $_;
	}
	@variables=@temparray;
	if ($query->param('addfield')) {
	    push @variables, 'biblio.title';
	}
	$self->{VARIABLES}=\@variables;
	$self->{COLUMNHEADERS}=\@variables;
	my $variables=join "\0", @variables;
	my $sth=$dbh->prepare("update reports set variables=? where id=?");
	$sth->execute($variables, $self->{ID});
    }
    my $variables=$self->{VARIABLES};
    print qq|
<h1>$self->{NAME}</h1>
<form>
<input type=hidden name=id value=$id>
<input type=hidden name=variables value=1>
<input type=hidden name=updatevariables value=1>
|;
    my $fieldcounter=0;
    foreach my $field (@$variables) {
	print "Field ".($fieldcounter+1).": <select name=variable-$fieldcounter>\n";
	foreach my $var (keys %$variabletable) {
	    my $f=$variabletable->{$var};
	    my $selected='';
	    if ($field eq $f) {
		$selected='selected';
	    }
	    print "<option value=$f $selected>$var\n";
	}
	print "</select> <input type=checkbox name=delvariable-$fieldcounter> Delete<p>\n";
	$fieldcounter++;
    }
    print qq|
    <input type=checkbox name=addfield> Add another field<p>
    <input type=submit value="Set Variables">
|;
}

sub htmlcriteria {
    my $self=shift;
    my @holder;
    print "<h1>$self->{NAME}</h1>\n";

    print "<form><input type=hidden name=id value=".$self->{ID}."><input type=hidden name=criteria value=1><input type=hidden name=setquery value=1>\n";
    print "<hr>\n";
    print "<table border=1 cellspacing=1 cellpadding=10 bgcolor=#cccccc><tr><td align=center>\n";
    my $queryconditions= [ @{$self->{QUERY}} ];
    unless ($#{$queryconditions}>=0) {
	print "Error: Empty criteria! ".$#{$queryconditions}."<p>\n";
	$queryconditions = [ 'CONTAINS', 'Title', 'Jack' ];
	$self->{QUERY}=[ @$queryconditions ]
    }
    my $elementcounter=$#{@{$self->{QUERY}}};
    foreach my $element (reverse @$queryconditions) {
    #while ( my $element = pop @$queryconditions) {
	if ($element eq 'AND') {
	    my $part1=pop @holder;
	    my $part2=pop @holder;
	    #push @holder, "($part1\nAND\n    $part2)";
	    push @holder, qq|
	    <table border=1 cellspacing=0 cellpadding=3 bgcolor=#99ffff>
	    <tr><td rowspan=2><select name=element-$elementcounter><option value=AND selected>AND<option value=OR>OR</select></td><td>$part1</td></tr>
	    <tr><td>$part2</td></tr>
	    </table>
	    |;
	} elsif ($element eq 'OR') {
	    my $part1=pop @holder;
	    my $part2=pop @holder;
	    #push @holder, "($part1\nOR\n    $part2)";
	    push @holder, qq|
	    <table border=1 cellspacing=0 cellpadding=3 bgcolor=#ff99ff>
	    <tr><td rowspan=2><select name=element-$elementcounter><option value=AND>AND<option value=OR selected>OR</select></td><td>$part1</td></tr>
	    <tr><td>$part2</td></tr>
	    </table>
	    |;
	} elsif ($element eq 'NOT') {
	} elsif ($OPERATORS->{$element}) {
	    my $part1=pop @holder;
	    my $part2=pop @holder;
	    $part1=variabledropdownbox(@$part1);
	    $part2=CGI->textfield(-name=>'element-'.$part2->[1], -default=>$part2->[0]);
	    #my $parent=$self->findparent($elementcounter);
	    my $notcheckbox='';
	    if ($self->{QUERY}->[$elementcounter-1] eq 'NOT') {
		$notcheckbox="<input type=checkbox name=element-".($elementcounter-1)." value=NOT checked> Not";
	    } else {
		$notcheckbox="<input type=checkbox name=addnot-$elementcounter> Not";
	    }
	    push @holder, "$part1 ".comparisondropdownbox($element,$elementcounter)." $part2<br>$notcheckbox<input type=checkbox name=del-$elementcounter value=Del> Del <input type=checkbox name=add-$elementcounter value=Add-$elementcounter> Add";
	} else {
	    push @holder, [$element, $elementcounter];
	}
	$elementcounter--;
    }
    my $criteriastring=join " ", @holder;
    print "$criteriastring\n";
    print "<p><input type=submit value='Set Query'>\n";
    print "</td></tr></table>\n";
    print "</form></body></html>\n";
}

sub ShowReverseNotation {
    my $self=shift;
    print "<h1>$self->{NAME}</h1>\n";
    my $queryconditions= [ @{$self->{QUERY}} ];
    print "<hr><h3>Reverse Notation Query</h3>\n";
    foreach (@$queryconditions) {
	if ($_ =~/\s/) {
	    print "\"$_\" ";
	} else {
	    print "$_ ";
	}
    }
    print qq|
<hr>
Reverse polish notation is read from right to left.  For example:
<p>
<code>AND CONTAINS Title egg CONTAINS Author John</code>
<p>
Means all records where <font color=#dd4444>Author CONTAINS John   AND  Title CONTAINS egg</font>
|;
}


sub querylist {
    my $self=shift;
    my $id=$self->{ID};
    my $dbh=C4::Context->dbh();
    my $sth=$dbh->prepare("select id,name from reports");
    $sth->execute;
    print "<ul>\n";
    while (my ($qid, $name) = $sth->fetchrow) {
	($name) || ($name='Unnamed query');
	print "<li><a href=reports.pl?id=$qid&querylist=1>$name</a>";
	if ($qid == $id) {
	    print "<--- currently selected query\n";
	} else {
	    print "\n";
	}
    }
    print "</ul>\n";
    print qq|
    
<form>
<input type=hidden name=querylist value=1>
<input type=hidden name=id value=$id>
Add a new query named <input name=newqueryname size=20>
<input type=submit value="Add">
</form>
|;
}

sub addquery {
    my $self=shift;
    my $newqueryname=shift;
    my $dbh=C4::Context->dbh();
    my $sth=$dbh->prepare("insert into reports (name, variables, criteria) values (?, ?, ?)");
    $sth->execute($newqueryname, "biblio.title\0bibliosubtitle.subtitle\0biblio.author", "CONTAINS\0Author\0Jack");
}


sub runquery {
    my $self=shift;
    print "<h1>$self->{NAME}</h1>\n";
    my %params=@_;
    my $showsql=$params{showsql};
    my $showtables=$params{showtables};
    my $dbh=C4::Context->dbh();
    my $validvariables=ValidVariables();
    my $variables=$self->{VARIABLES};

    my $queryconditions= [ @{$self->{QUERY}} ];
    my $tables;
    my @holder;
    while ( my $element = pop @$queryconditions) {
	if ($element eq 'AND') {
	    my $part1=pop @holder;
	    my $part2=pop @holder;
	    push @holder, "($part1\nAND\n    $part2)";
	} elsif ($element eq 'OR') {
	    my $part1=pop @holder;
	    my $part2=pop @holder;
	    push @holder, "($part1\nOR\n    $part2)";
	} elsif ($element eq 'NOT') {
	    my $part1=pop @holder;
	    push @holder, "!($part1)";
	} elsif ($element eq 'CONTAINS') {
	    my $part1=pop @holder;
	    my $part2=pop @holder;
	    $part2=$dbh->quote($part2);
	    $part2=~s/^\'//;
	    $part2=~s/\'$//;
	    if ($part1 =~/ARRAY/) {
		my @subparts;
		foreach my $subpart (@$part1) {
		    push @subparts, qq|( $subpart LIKE \"$part2%" OR $subpart LIKE "% $part2%" )|;
		}
		push @holder, '( '. join(' OR ', @subparts) . ' )';
	    } else {
		push @holder, qq|( $part1 LIKE \"$part2%" OR $part1 LIKE "% $part2%" )|;
	    }
	} elsif ($element eq 'EQUALS') {
	    my $part1=pop @holder;
	    my $part2=pop @holder;
	    push @holder, "$part1=".$dbh->quote($part2);
	} elsif ($element eq 'GT') {
	    my $part1=pop @holder;
	    my $part2=pop @holder;
	    push @holder, "$part1>".$dbh->quote($part2);
	} elsif ($element eq 'LT') {
	    my $part1=pop @holder;
	    my $part2=pop @holder;
	    push @holder, "$part1<".$dbh->quote($part2);
	} elsif ($element eq 'GE') {
	    my $part1=pop @holder;
	    my $part2=pop @holder;
	    push @holder, "$part1>=".$dbh->quote($part2);
	} elsif ($element eq 'LE') {
	    my $part1=pop @holder;
	    my $part2=pop @holder;
	    push @holder, "$part1<=".$dbh->quote($part2);
	} else {
	    if ($validvariables->{$element}->{tables}) {
		foreach my $table (@{$validvariables->{$element}->{tables}}) {
		    $tables->{$table}=1;
		}
		$element=$validvariables->{$element}->{field};
	    }
	    push @holder, $element;
	}
    }
    foreach my $var (@$variables) {
	$var=~/^(.*)\./;
	$tables->{$1}=1;
    }
    my $variablestring=join ", ", @$variables;
    my $tablestring=join ", ", keys %$tables;
    if ($showtables) {
	print "<h2>Tables used in this query</h2>\n";
	foreach (keys %$tables) {
	    print "$_<br>\n";
	}
	return;
    }
    my $criteriastring=join " ", @holder;
    my $tc=keys %$tables;
    if (keys %$tables > 1) {
	my $tablesjoined;
	my $tablejoincriteria=tablejoincriteria();
TABLE:
	foreach my $table (keys %$tables) {
	    (next) if ($tablesjoined->{$table});
	    foreach my $tjc (@$tablejoincriteria) {
		if ($tjc->{table1} eq $table) {
		    if ($tables->{$tjc->{table2}}) {
			$tablesjoined->{$table}=1;
			$tablesjoined->{$tjc->{table2}}=1;
			$criteriastring.="\nAND\n    $table.".$tjc->{variable1}."=".$tjc->{table2}.".".$tjc->{variable2}."\n";
			next TABLE;
		    }
		} elsif ($tjc->{table2} eq $table) {
		}
	    }
	}
	foreach my $table1 (keys %$tables) {
	    foreach my $table2 (keys %$tables) {
		foreach my $tjc (@$tablejoincriteria) {
		    if ($tjc->{table1} eq $table1 and $tjc->{table2} eq $table2 and $tjc->{mandatory}
		     or $tjc->{table2} eq $table1 and $tjc->{table1} eq $table2 and $tjc->{mandatory}) {
			my $checkstring1="\nAND\n    $table1.".$tjc->{variable1}."=$table2.".$tjc->{variable2}."\n";
			my $checkstring2="\nAND\n    $table2.".$tjc->{variable2}."=$table1.".$tjc->{variable1}."\n";
			unless ($criteriastring=~/$checkstring1/ or $criteriastring=~/$checkstring2/) {
			    $criteriastring.=$checkstring1;
			}
		    }
		}
	    }
	}
    }
    my $querystring= qq|
select
    $variablestring
from
    $tablestring
where
    $criteriastring
limit 100
|;
    if ($showsql) {
	print "<hr><h3>SQL Query String</h3>\n";
	print "<pre>$querystring</pre><hr>\n";
	return;
    }
    my $sth=$dbh->prepare("EXPLAIN $querystring");
    $sth->execute;
    my $rows=1;
    while (my $results=$sth->fetchrow_hashref) {
	$rows*=$results->{rows};
    }
    print "ROWS TO JOIN: $rows<br>\n";

    $sth=$dbh->prepare($querystring);
    $sth->execute;
    print "<table border=1 cellspacing=0>\n";
    print "<tr><th>\n";
    print join "\n</th><th>\n", @{$self->{COLUMNHEADERS}};
    print "</th></tr>\n";
    while (my @result=$sth->fetchrow_array) {
	print "<tr><td>\n";
	print join "\n</td><td>\n", @result;
	print "\n</td></tr>\n";
    }
    print "</table>\n";
}

sub tablejoincriteria {
    return 
    [
    	{ 'table1' => 'biblio', 'table2' => 'biblioitems', 'variable1' => 'biblionumber', 'variable2' => 'biblionumber', 'mandatory' => 1 },
    	{ 'table1' => 'biblio', 'table2' => 'items', 'variable1' => 'biblionumber', 'variable2' => 'biblionumber' },
    	{ 'table1' => 'biblio', 'table2' => 'bibliosubtitle', 'variable1' => 'biblionumber', 'variable2' => 'biblionumber' },
    	{ 'table1' => 'biblio', 'table2' => 'bibliosubject', 'variable1' => 'biblionumber', 'variable2' => 'biblionumber' },
    	{ 'table1' => 'biblioitems', 'table2' => 'items', 'variable1' => 'biblioitemnumber', 'variable2' => 'biblioitemnumber', 'mandatory' => 1 },
    	{ 'table1' => 'itemtypes', 'table2' => 'biblioitems', 'variable1' => 'itemtype', 'variable2' => 'itemtype' },
    	{ 'table1' => 'bibliosubtitle', 'table2' => 'biblio', 'variable1' => 'biblionumber', 'variable2' => 'biblionumber' },
    	{ 'table1' => 'bibliosubject', 'table2' => 'biblio', 'variable1' => 'biblionumber', 'variable2' => 'biblionumber' },
    	{ 'table1' => 'issues', 'table2' => 'items', 'variable1' => 'itemnumber', 'variable2' => 'itemnumber', 'mandatory' => 1 },
    	{ 'table1' => 'items', 'table2' => 'issues', 'variable1' => 'itemnumber', 'variable2' => 'itemnumber' },
    	{ 'table1' => 'borrowers', 'table2' => 'issues', 'variable1' => 'borrowernumber', 'variable2' => 'borrowernumber' },
    	{ 'table1' => 'issues', 'table2' => 'borrowers', 'variable1' => 'borrowernumber', 'variable2' => 'borrowernumber', 'mandatory' => 1 },
    ];
}

sub ValidVariables {
    my $validvariables= {
    	'Keyword' => { 'field' => 'KEYWORD', 'tables' => [ 'biblio', 'biblioitems', 'bibliosubject', 'bibliosubtitle' ] },
    	'Title' => { 'field' => [ 'biblio.title', 'bibliosubtitle.subtitle'], 'tables' => [ 'biblio', 'bibliosubtitle' ] },
    	'Author' => { 'field' => 'biblio.author', 'tables' => [ 'biblio' ] },
    	'Copyright Date' => { 'field' => 'biblio.copyrightdate', 'tables' => [ 'biblio' ] },
    	'Publication Year' => { 'field' => 'biblioitems.publicationyear', 'tables' => [ 'biblioitems' ] },
    	'Item Type' => { 'field' => 'biblioitems.itemtype', 'tables' => [ 'biblioitems' ] },
    	'ISBN' => { 'field' => 'biblioitems.isbn', 'tables' => [ 'biblioitems' ] },
    	'Dewey' => { 'field' => 'biblioitems.dewey', 'tables' => [ 'biblioitems' ] },
    	'Barcode' => { 'field' => 'items.barcode', 'tables' => [ 'items' ] },
	'Lost Item' => { 'field' => 'items.itemlost', 'tables' => [ 'items' ] },
    	'Issued To' => { 'field' => 'issues.borrowernumber', 'tables' => [ 'issues' ] },
    	'Due Date' => { 'field' => 'issues.date_due', 'tables' => [ 'issues' ] },
    };
    return $validvariables;
}

1;
