package Reports;

=head1 NAME

Reports - Perl extension for generating generic reports in Koha

=cut

use strict;
use CGI;
use C4::Context;
use Cache::FileCache;

=head1 VERSION

  $Id:$

=cut

=head1 DESCRIPTION

Module for querying and stocking Virtual Bookshelves

   1. can contain a list of items, a list of biblioitems, or a list of biblios
   2. can have an arbitrary name, and will have a unique numerical identifier
   3. will have arbitrary metadata (properties) associated with it
	  * Sharing information (private, only visible by the owner of the
	    Reports; shared with a group of patrons; public, viewable by anybody)
	  * Special circulation rules - Do not return to home branch, do not
	    circulate, reduced loan time (ie 3 day loan)
	  * Search query term - if the Reports is the result of a query, the
	    query itself can be stored with the list of books that resulted
          * Creation date - useful for 'retiring' a stale cached query result
          * Access information - who has "write" or "read" access to the Reports.
	  * Searchable - If a patron can perform a search query on the contents
	    of this Reports


Patrons typically will only use "biblioitem" bookshelves, and will not need to
be presented with the differences between biblioitem and item bookshelves.


Some uses for VirtualBookshelves

   1. Cache search results for faster response on popular searches
   2. Name search results so that patrons can pull up saved searches
   3. Creation of sub-collections within a library or branch
   4. replacing "itemtypes" field... this would allow an individual item to be
	a member of more than one itemtype
   5. store a patron's reading record (if he chooses to store such data)
   6. store a patron's "To be read" list
   7. A teacher of a course could add a list of books to a Reports for his course
	and ask that those items be marked non-circulating so students always
	have access to them at the library.
	  * The teacher creates the list of materials that she wants to be
	    non-circulating (or reduced to 3-day loan) and marks them as such
	  * A librarian receives a notice that a Reports requires her attention.
	    He can pull up a list of the contents of the Reports, the owner of
	    the Reports, and the reason the owner is requesting this change in
	    circulation rules. The librarian can approve or deny the request.
	  * Optionally, create an access flag that grants teachers the right to
	    put items on modified circulation shelves without librarian
	    intervention.


=cut


my $OPERATORS= {
    'EQUALS' => { 'name' => 'is equal to', 'order' => 1 },
    'NOTEQUALS' => { 'name' => 'is not equal to', 'order' => 2  },
    'CONTAINS' => { 'name' => 'contains', 'order' => 3  },
    'NOTCONTAINS' => { 'name' => 'does not contain', 'order' => 4  },
    'GT' => { 'name' => '&gt;', 'order' => 5  },
    'LT' => { 'name' => '&lt;', 'order' => 6  },
    'GE' => { 'name' => '&ge;', 'order' => 7  },
    'LE' => { 'name' => '&le;', 'order' => 8  },
};



my $QC= [ 'AND', 'AND', 'OR', 'CONTAINS', 'biblio.title', 'cat', 'CONTAINS', 'biblio.title', 'dog', 'OR', 'CONTAINS', 'biblio.author', 'Hall', 'OR', 'CONTAINS', 'biblio.author', 'Berger', 'NOTCONTAINS', 'biblio.author', 'gold', 'AND', 'GT', 'items.barcode', '10000', 'LT', 'items.barcode', '20000' ];

=head1 METHODS

=head2 C<new()>

Base constructor for the class.

  my $Reports=Reports->new(56);
      will load bookReports 56.
  my $Reports=Reports->new(-name => 'Fiction');
  my $Reports=Reports->new('Fiction');
      will load the internal 'Fiction' Reports
  my $Reports=Reports->new('Favourite Books', 'sjohnson');
  my $Reports=Reports->new(-name => 'Favourite Books', -owner => 'sjohnson');
      will load sjohnson's "Favourite Books" bookReports
  
  Any of the last four invocations will create a new Reports with the name and
  owner given if one doesn't already exist.


=cut

sub new {
    my $self = {};
    $self->{ID}	= undef;
    $self->{NAME}=undef;
    $self->{QUERY}=undef;
    $self->{VARIABLES}=undef;
    $self->{COLUMNHEADERS}=undef;

    if (@_) {
	my $dbh=C4::Context->dbh();
	shift;
	if ($#_ == 0) {
	    $self->{ID}=shift;
	    # load attributes of Reports #ID
	} elsif ($#_) {
	    my ($name,$query,$variables);
	    if ($_[0] =~/^-/) {
		my %params=@_;
		$name=$params{name};
		$query=$params{query};
		$variables=$params{variables};
	    } else {
		$name=shift;
		$query=shift;
		$variables=shift;
	    }
	}
    }
    bless($self);
    return $self;
}

sub variabledropdownbox {
    my $variable=shift;
    my $validvariables=ValidVariables();
    my $element="<select name=whatever>\n";
    foreach my $validvariable (sort keys %$validvariables) {
	my $selected='';
	$validvariable eq $variable and $selected=' selected';
	$element.="<option value=$validvariable$selected>$validvariable\n";
    }
    $element.="</select>\n";
    return $element;
}

sub comparisondropdownbox {
    my $comparison=shift;
    my $element="<select name=comparison-45>\n";
    foreach my $op (sort { $OPERATORS->{$a}->{order} <=> $OPERATORS->{$b}->{order} } keys %$OPERATORS) {
	my $selected='';
	warn "$op - $comparison\n";
	$op eq $comparison and $selected=' selected';
	my $opname=$OPERATORS->{$op}->{name};
	$element.="<option value=$op$selected>$opname\n";
    }
    $element.="</select>\n";
    return $element;
}

sub htmlcriteria {
    my $self=shift;
    my @holder;
    print "<html><body><form>\n";
    my $queryconditions= [ @$QC ];
    my $elementcounter=0;
    while ( my $element = pop @$queryconditions) {
	if ($element eq 'AND') {
	    my $part1=pop @holder;
	    my $part2=pop @holder;
	    #push @holder, "($part1\nAND\n    $part2)";
	    push @holder, qq|
	    <table border=1 cellspacing=0 cellpadding=3 bgcolor=#99ffff>
	    <tr><td rowspan=2><select name=select-asdf><option value=AND selected>AND<option value=OR>OR</select></td><td>$part1</td></tr>
	    <tr><td>$part2</td></tr>
	    </table>
	    |;
	} elsif ($element eq 'OR') {
	    my $part1=pop @holder;
	    my $part2=pop @holder;
	    #push @holder, "($part1\nOR\n    $part2)";
	    push @holder, qq|
	    <table border=1 cellspacing=0 cellpadding=3 bgcolor=#ff99ff>
	    <tr><td rowspan=2><select name=select-asdf><option value=AND>AND<option value=OR selected>OR</select></td><td>$part1</td></tr>
	    <tr><td>$part2</td></tr>
	    </table>
	    |;
	} elsif ($element eq 'CONTAINS') {
	    my $part1=variabledropdownbox(pop @holder);
	    my $part2=CGI->textfield(-name=>'whateveragain', -default=>pop @holder);
	    push @holder, "$part1 ".comparisondropdownbox($element)." $part2";
	} elsif ($element eq 'NOTCONTAINS') {
	    my $part1=variabledropdownbox(pop @holder);
	    my $part2=CGI->textfield(-name=>'whateveragain', -default=>pop @holder);
	    push @holder, "$part1 ".comparisondropdownbox($element)." $part2";
	} elsif ($element eq 'EQUALS') {
	    my $part1=variabledropdownbox(pop @holder);
	    my $part2=CGI->textfield(-name=>'whateveragain', -default=>pop @holder);
	    push @holder, "$part1 ".comparisondropdownbox($element)." $part2";
	} elsif ($element eq 'NOTEQUALS') {
	    my $part1=variabledropdownbox(pop @holder);
	    my $part2=CGI->textfield(-name=>'whateveragain', -default=>pop @holder);
	    push @holder, "$part1 ".comparisondropdownbox($element)." $part2";
	} elsif ($element eq 'GT') {
	    my $part1=variabledropdownbox(pop @holder);
	    my $part2=CGI->textfield(-name=>'whateveragain', -default=>pop @holder);
	    push @holder, "$part1 ".comparisondropdownbox($element)." $part2";
	} elsif ($element eq 'LT') {
	    my $part1=variabledropdownbox(pop @holder);
	    my $part2=CGI->textfield(-name=>'whateveragain', -default=>pop @holder);
	    push @holder, "$part1 ".comparisondropdownbox($element)." $part2";
	} else {
	    push @holder, [$element, $elementcounter];
	}
	$elementcounter++;
    }
    my $criteriastring=join " ", @holder;
    print "$criteriastring\n";
    print "</form></body></html>\n";
}

sub runquery {
    my $self=shift;
    my $dbh=C4::Context->dbh();
    my $validvariables=ValidVariables();
    my $variables = [ 'biblio.biblionumber', 'biblio.title', 'biblio.author', 'biblioitems.itemtype', 'biblioitems.isbn', 'biblioitems.dewey', 'items.barcode' ];
    my $queryconditions= [ @$QC ];
    print "<hr><h3>Reverse Notation Query</h3>\n";
    print join(' ', @$queryconditions);
    my $tables;
    my @holder;
    while ( my $element = pop @$queryconditions) {
	if ($element eq 'AND') {
	    my $part1=pop @holder;
	    my $part2=pop @holder;
	    push @holder, "($part1\nAND\n    $part2)";
	} elsif ($element eq 'OR') {
	    my $part1=pop @holder;
	    my $part2=pop @holder;
	    push @holder, "($part1\nOR\n    $part2)";
	} elsif ($element eq 'CONTAINS') {
	    my $part1=pop @holder;
	    my $part2=pop @holder;
	    $part2=$dbh->quote($part2);
	    $part2=~s/^\'//;
	    $part2=~s/\'$//;
	    push @holder, qq|( $part1 LIKE \"$part2%" OR $part1 LIKE "% $part2%" )|;
	} elsif ($element eq 'NOTCONTAINS') {
	    my $part1=pop @holder;
	    my $part2=pop @holder;
	    $part2=$dbh->quote($part2);
	    $part2=~s/^\'//;
	    $part2=~s/\'$//;
	    push @holder, qq|( !(  $part1 LIKE \"$part2%" OR $part1 LIKE "% $part2%" ) )|;
	} elsif ($element eq 'EQUALS') {
	    my $part1=pop @holder;
	    my $part2=pop @holder;
	    push @holder, "$part1=".$dbh->quote($part2);
	} elsif ($element eq 'GT') {
	    my $part1=pop @holder;
	    my $part2=pop @holder;
	    push @holder, "$part1>".$dbh->quote($part2);
	} elsif ($element eq 'LT') {
	    my $part1=pop @holder;
	    my $part2=pop @holder;
	    push @holder, "$part1<".$dbh->quote($part2);
	} else {
	    if ($validvariables->{$element}) {
		$tables->{$validvariables->{$element}}=1;
	    }
	    push @holder, $element;
	}
    }
    foreach my $var (@$variables) {
	$tables->{$validvariables->{$var}}=1;
    }
    my $variablestring=join ", ", @$variables;
    my $tablestring=join ", ", keys %$tables;
    my $criteriastring=join " ", @holder;
    my $tc=keys %$tables;
    if (keys %$tables > 1) {
	my $tablesjoined;
	my $tablejoincriteria=tablejoincriteria();
TABLE:
	foreach my $table (keys %$tables) {
	    (next) if ($tablesjoined->{$table});
	    foreach my $tjc (@$tablejoincriteria) {
		if ($tjc->{table1} eq $table) {
		    if ($tables->{$tjc->{table2}}) {
			$tablesjoined->{$table}=1;
			$tablesjoined->{$tjc->{table2}}=1;
			$criteriastring.="\nAND\n    $table.".$tjc->{variable1}."=".$tjc->{table2}.".".$tjc->{variable2}."\n";
			next TABLE;
		    }
		} elsif ($tjc->{table2} eq $table) {
		}
	    }
	}
    }
    my $querystring= qq|
select
    $variablestring
from
    $tablestring
where
    $criteriastring
|;
    print "<hr><h3>SQL Query String</h3>\n";
    print "<pre>$querystring</pre><hr>\n";
    #my $sth=$dbh->prepare("EXPLAIN $querystring");
    #$sth->execute;
    #while (my @result=$sth->fetchrow_array) {
#	print join " ", @result;
#	print "<br>\n";
#    }
    my $sth=$dbh->prepare($querystring);
    $sth->execute;
    print "<table border=1 cellspacing=0>\n";
    print "<tr><th>\n";
    print join "\n</th><th>\n", @$variables;
    print "</th></tr>\n";
    while (my @result=$sth->fetchrow_array) {
	print "<tr><td>\n";
	print join "\n</td><td>\n", @result;
	print "\n</td></tr>\n";
    }
    print "</table>\n";
}

sub tablejoincriteria {
    return 
    [
    	{ 'table1' => 'biblio', 'table2' => 'biblioitems', 'variable1' => 'biblionumber', 'variable2' => 'biblionumber' },
    	{ 'table1' => 'biblioitems', 'table2' => 'items', 'variable1' => 'biblionumber', 'variable2' => 'biblionumber' },
    	{ 'table1' => 'biblio', 'table2' => 'items', 'variable1' => 'biblionumber', 'variable2' => 'biblionumber' },
    	{ 'table1' => 'biblio', 'table2' => 'bibliosubtitle', 'variable1' => 'biblionumber', 'variable2' => 'biblionumber' },
    	{ 'table1' => 'biblio', 'table2' => 'bibliosubject', 'variable1' => 'biblionumber', 'variable2' => 'biblionumber' },
    ]
}

sub ValidVariables {
    my $variables= [
    	'biblio.biblionumber',
	'biblio.title',
	'biblio.author',
	'biblio.copyrightdate',
	'biblioitems.itemtype',
	'biblioitems.isbn',
	'biblioitems.dewey',
	'items.barcode',
    ];
    my $validvariables;
    foreach (@$variables) {
	m#(.*)\.#;
	$validvariables->{$_}=$1
    }
    return $validvariables;
}

1;
